#include "jarray.h"
#include "jengine.h"
#include <string>
#include <cassert>
#include <cstring>
#include <cstdlib>
#include <iostream>

const bool TRACEMEM=false;

jarray::jarray():hdr(NULL) {
};

void jarray::grab() const {
  if (hdr) hdr->refcnt++;
};

void jarray::release() {
  if (hdr) {
    hdr->refcnt--;
    if (hdr->refcnt==0) {
      if (hdr->flag&OWNED_BY_J) {
	assert((hdr->flag&OWNED_BY_JPLUS)==0); // won't pass our own memory
	hdr->flag^=OWNED_BY_J;
	hdr->refcnt++;
      } else { // free memory
	if (hdr->flag&OWNED_BY_JPLUS) {
	  MS* ms=(MS*)hdr-1;
	  free(ms);
	} else {
	  hdr->refcnt++;
	  // even this is unnecessary as jtga will register allocations in
	  // tempstack and free them automatically.
	  // jengine::FR(hdr); // this deallocates memory in J memory pool
	};
      };
    };
  };    
};

jarray::jarray(jengine *je_,void *hdr_):hdr((header*) hdr_),je(je_) {
  assert(hdr);
  assert(hdr->refcnt);
  grab();
};

jarray::I jarray::esize(elementType type) {
  switch(type) {
  case T_B01:
    return sizeof(B);
  case T_LIT:
    return sizeof(C);
  case T_INT:
    return sizeof(I);
  case T_FL:
    return sizeof(D);
  case T_CMPX:
    return sizeof(Z);
  default:
    assert(false); // unsupported type
    return 0;
  };
};

jarray::I jarray::esize() const {
  assert(hdr);
  return esize((elementType)hdr->type);
};

bool jarray::allocate(jengine * je_, jarray::elementType type, 
		      const I rank, const I* shape) {
  je=je_;
  int n=1;
  for(int i=0;i<rank;i++) n*=shape[i];
  if (je) {
    // when J engine is initialized, we cooperate with J
    // on memory allocation (and deallocation)
    hdr=(header*)(je->GA(type,n,rank,shape));
    assert(hdr);
  } else {
    // but when there is no J engine around we manage the
    // memory ourselves
    int e=esize(type); // element size
    int k=(7+rank)*sizeof(I);
    int maxbytes=n*e;
    MS* ms=(MS*)malloc(sizeof(MS)+k+maxbytes);
    ms->a=0;
    ms->j=11;
    ms->mflag=0;
    hdr=(header*)(ms+1);
    hdr->offset=k;
    hdr->flag=OWNED_BY_JPLUS;
    hdr->maxbytes=maxbytes;
    hdr->type=type;
    hdr->refcnt=1;
    hdr->n=n;
    hdr->rank=rank;
    // copy the shape
    for(int i=0;i<rank;i++)
      ((I*)hdr)[7+i]=shape[i];
  };
  return true;
};

jarray::jarray(jengine *je_, elementType type, I rank, I* shape):hdr(NULL),je(NULL) {
  allocate(je_, type, rank, shape);
};

jarray::jarray(jengine *je_, elementType type, const std::vector<I>& shape):hdr(NULL),je(NULL) {
  allocate(je_, type, shape.size(), &shape[0]);
};

jarray::jarray(jengine *je_, const std::string &str):hdr(NULL),je(NULL) {
  I shape[]={str.size()};
  allocate(je_, T_LIT, 1, shape);
  str.copy((C*)data(),shape[0],0);
};

jarray::jarray(jengine *je_, std::istream &in):hdr(NULL),je(NULL) {
  // since we now cooperate with J on memory allocation msize and
  // ms are, strictly speaking, unnecessary
  // but we keep them for the sake of on-disk compatibility
  // with previous version
  int msize;
  in.read((char*)&msize,sizeof(msize));
  MS ms;
  in.read((char*)&ms,sizeof(ms));
  I offset,flag,maxbytes,type,refcnt,n,rank;
  in.read((char*)&offset,sizeof(offset));
  in.read((char*)&flag,sizeof(flag));
  in.read((char*)&maxbytes,sizeof(maxbytes));
  in.read((char*)&type,sizeof(type));
  in.read((char*)&refcnt,sizeof(refcnt));
  in.read((char*)&n,sizeof(n));
  in.read((char*)&rank,sizeof(rank));
  I shape[rank];
  in.read((char*)&shape,rank*sizeof(I));
  allocate(je_, (elementType)type,rank,shape);
  int dsize=size()*esize();
  in.read((char*)data(),dsize); // read data
//   int msize;
//   in.read((char*)&msize,sizeof(msize));
//   MS* ms=(MS*)malloc(msize);
//   in.read((char*)ms,msize);
//   hdr=(header*)(ms+1);
//   ms->a=0;
//   ms->j=11;
//   ms->mflag=0;
//   hdr->refcnt=1;
//   hdr->offset=(7+rank())*sizeof(I); // when read data follow immediately
};

bool jarray::write(std::ostream& out) {
  int hsize=sizeof(MS)+(7+rank())*sizeof(I);
  int dsize=size()*esize();
  int msize=hsize+dsize;
  MS* ms=(MS*)hdr-1;
  out.write((char*)&msize,sizeof(msize));
  // two writes to account for arrays possibly discontinuous in memory
  out.write((char*)ms,hsize); // write header
  out.write((char*)data(),dsize); // write data
};

/**
 * Add this array (both shape and values) to the hash.
 */
void jarray::addhash(SHA1 &sha1) {
  sha1.Input((char*)&(hdr->type),sizeof(I));       // type
  sha1.Input((char*)&(hdr->rank),sizeof(I));       // rank
  sha1.Input((char*)shape(),hdr->rank*sizeof(I));  // shape
  sha1.Input((char*)data(),size()*esize());        // data
};

bool jarray::operator== ( const jarray& rhs ) const {
  if (type()!=rhs.type()) return false;
  if (rank()!=rhs.rank()) return false;
  for (int i=0;i<rank();i++)
    if (shape()[i]!=rhs.shape()[i]) return false;
  if (memcmp(data(), rhs.data(), size()*esize())) return false;
  return true;
};


jarray::jarray(const jarray &other) {
  other.grab();
  hdr=other.hdr; // shallow copy
  je=other.getEngine(); // inherit the memory allocator
};

jarray& jarray::assign(const jarray&other) {
  release();
  other.grab();
  hdr=other.hdr;
  je=other.getEngine(); // inherit the memory allocator  
};

jarray::~jarray() {
  release();
  if (TRACEMEM)
    std::cerr << "array ref freed "<< (I)hdr 
	      <<", refcount=" << (hdr?hdr->refcnt:0) << std::endl; 
};

template<> int jarray::get(std::string& v) { // gets the whole array
  if (hdr->type!=T_LIT) return ERR_CONV;
  std::string r((char*)data(),size());
  r.swap(v);
  return 0;
};

// right now just dumps ravel, but ideally should print arrays exactly as
// J interpreter does. TODO.
std::ostream& operator<< (std::ostream& stream, const jarray& a) {
  jarray::I *s=a.shape();
  int n=1;
  for(int i=0;i<a.rank();i++) n*=s[i];
  for(int i=0;i<n;i++) {
    switch (a.type()) {
    case jarray::T_LIT:
      char vc;
      a.get(vc,i);
      stream << vc;
      break;
    case jarray::T_B01:
    case jarray::T_INT:
      int vi;
      a.get(vi,i);
      stream << vi;
      break;
    case jarray::T_FL:
      double vd;
      a.get(vd,i);
      stream << vd;
      break;
    case jarray::T_CMPX:
      std::complex<double> vz;
      a.get(vz,i);
      stream << vz;
    };
    if ((a.type()!=jarray::T_LIT) && (i<n-1)) stream << ' ';
  };
  return stream;
};
