#include "config.h"
#include "yacts.h"
#include <set>
#include <string>
#include <vector>
#include <sstream>
#include <iostream>
#include <fstream>
#include "sha1.h"

// components of CVODE
#include <cvode/cvode.h>
#include <cvode/cvode_spgmr.h>
#include <nvector/nvector_serial.h>
#include <sundials/sundials_nvector.h>

using namespace std;

bool PROGRESS=false; // global variable for enabling progress display

/** initialized yacts, loads yacts script */
yacts::yacts():jplus(),outName("OUT") {
  cvode_mem = NULL;
  nvs = NULL;
};

bool yacts::init(std::istream &script) {
  jplus::init(script);
};

bool yacts::setOut(string newOut) {
  if (newOut.size()==0) return false;
  bool ok=((newOut[0]>='A')&&(newOut[0]<='Z')) ||
    ((newOut[0]>='a')&&(newOut[0]<='z'));
  for (int i=1;(i<newOut.size())&&ok;i++)
    ok=((newOut[i]>='A')&&(newOut[i]<='Z'))||
      ((newOut[i]>='a')&&(newOut[i]<='z'))||
      ((newOut[i]>='0')&&(newOut[i]<='9'));
  if (!ok) return false;
  outName=newOut;
  return true;
};

string yacts::getTrajectoryFilenameBase(string prefix) {
  // obtain the program for computing the important variables
  std::set<std::string> varset;
  varset.insert("T");
  varset.insert("S");
  varset.insert("dSdT");
  vector<string> program=getProgram(varset);
  // hash it
  SHA1 sha1;
  for(int i=0;i<program.size();i++)
    sha1.Input(program[i].c_str(),program[i].size());
  // form the filename
  if (prefix.size()==0) 
    return sha1tostring(sha1);
  else
    return prefix+"_"+sha1tostring(sha1);
};

void setS(yacts *This, N_Vector s) {
  int ns=NV_LENGTH_S(s);
  jarray oldS=This->get("S");
  vector<jarray::I> shape;
  oldS.shape(shape); // get the shape of the old S
  assert(oldS.size()==ns);
  jarray_of_type<double> js(This,shape); // create S, preserving the shape
  for(int i=0;i<ns;i++) js[i]=NV_Ith_S(s,i); // TODO: think about avoiding copy
  This->set("S",js); // set current state in J+
};

/** This function is called by CVODE to compute the time derivative*/
int rhs(realtype t, N_Vector s, N_Vector dsdt, void *tThis) {
  yacts *This=(yacts*)tThis;
  jarray_of_type<double> jt(This);
  jt[0]=t;
  This->set("T",jt); // set current time in J+
  int ns=NV_LENGTH_S(s); // number of equations as CVODE knows
  setS(This, s); // set current state
  jarray_of_type<double> jdsdt=This->get("dSdT"); // compute derivative
  assert(jdsdt.size()==ns);
  for(int i=0;i<ns;i++) NV_Ith_S(dsdt,i)=jdsdt[i];
  jarray_of_type<double> js=This->get("S"); // compute S (it should be already)
  for(int i=0;i<ns;i++) NV_Ith_S(s,i)=js[i];
  return 0;
};

bool yacts::saveFrame() {
  assert(trjfile::good());
  jarray_of_type<double> jtnext=get("T"); // compute next time
  jarray_of_type<double> jt=jengine::get("T"); // first version of T, curr time
  assert(jt.rank()==0); // assert the time is scalar
  jarray_of_type<double> js=get("S"); // current state
  return trjfile::saveFrame(js.size(), jt[0], &js[0]);
};

bool yacts::loadFrame() {
  assert(trjfile::good());
  if (!trjfile::hasNextFrame()) return false;
  jarray_of_type<double> jt(this);
  jarray_of_type<double> js(this,trjfile::getNEQ());
  if (trjfile::loadFrame(jt[0], &js[0])) {
    set("T",jt);
    set("S",js);
    return true;
  } else return false;
};

bool yacts::hasNextFrameStored() {
  return trjfile::hasNextFrame();
};

int yacts::size() {
  return trjfile::size();
};

bool yacts::setFrame(int iframe)  {
  return trjfile::toFrame(iframe);
};

bool yacts::initTrajectory(string prefix) {
  string trjname=getTrajectoryFilenameBase(prefix)+".trj";
  trjfile::open(trjname);
  if (!trjfile::hasNextFrame()) {
    // file is new, we need to create and save the initial frame
    saveFrame();
  } else {
    // read the last frame
    trjfile::toFrame(-1);
    loadFrame();
  };

  // now we have the first frame inside J+
  // pull it out and feed to the solver
  jarray_of_type<double> jtnext=get("T"); // compute next time
  jarray_of_type<double> jt=jengine::get("T"); // first version of T, curr time
  assert(jt.rank()==0); // assert the time is scalar

  jarray_of_type<double> js=get("S");

  nvs = N_VNew_Serial(neq); assert(nvs);
  for(int i=0;i<js.size();i++) NV_Ith_S(((N_Vector)nvs),i)=js[i];
  cvode_mem = CVodeCreate(CV_BDF, CV_NEWTON);
  int flag;
  realtype abstol=1.0e-5; /*abstol*/
#ifdef HAVE_CVODEMALLOC
  flag = CVodeMalloc(cvode_mem, rhs, jt[0], (N_Vector)nvs, 
		     CV_SS,1.0e-3/*reltol*/,&abstol);  assert(!flag);
  // pass pointer to "this" object for rhs
  flag = CVodeSetFdata(cvode_mem, this); assert(!flag);
#elif HAVE_CVODEINIT
  flag = CVodeInit(cvode_mem, rhs, jt[0], (N_Vector)nvs); assert(!flag);
  flag = CVodeSStolerances(cvode_mem,1.0e-3 ,abstol); assert(!flag);
  // pass pointer to "this" object for rhs
  flag = CVodeSetUserData(cvode_mem, this); assert(!flag);
#else
#error Unknown version of SUNDIALS API.
#endif
  // specify the solver and subspace dimension
  flag = CVSpgmr(cvode_mem,PREC_NONE,300);

  // now we are ready to integrate
};

bool yacts::nextFrame() {
  // just load from trajectory if we can
  if (hasNextFrameStored()) return loadFrame();

  // first check what script wants from us
  jarray_of_type<double> jtnext=get("T"); // compute next time
  jarray_of_type<double> jt=jengine::get("T"); // first version of T, curr time
  assert(jtnext.rank()==0);
  assert(jt.rank()==0);
  if (tol_eq(jtnext[0],jt[0])) return false; // the end, no more frames

  // integrate until the next frame
  realtype t;
  int flag;
  flag = CVode(cvode_mem, jtnext[0], (N_Vector)nvs, &t, CV_NORMAL);
  assert(!flag);
  jt[0]=t;
  set("T",jt); // set current time in J+
  setS(this, (N_Vector)nvs); // set current state in J+
  saveFrame(); // save frame  
  
  return true;
};

std::string yacts::process() {
  // the current frame is loaded, we just evaluate 'OUT'
  jarray out=get(outName);
  assert(out.type()==jarray::T_LIT);
  string res;
  out.get(res);
  return res;
};

/** frees yacts memory */
yacts::~yacts() {
  if (cvode_mem) CVodeFree(&cvode_mem);
  if (nvs) N_VDestroy_Serial((N_Vector)nvs);
};

typedef jengine::monad monad;
typedef jengine::dyad dyad;

// prototypes for library functions, defined in l_XXX.cpp files
jarray fft(jarray x, jarray y);                      // l_fftw.cpp
jarray fft(jarray y);                                // l_fftw.cpp
jarray fftc2r(jarray x, jarray y);                   // l_fftw.cpp
jarray fftr2c(jarray y);                             // l_fftw.cpp
jarray cuba(monad um, dyad ud, jarray x, jarray y); // l_cuhre.cpp
jarray cuba(monad um, dyad ud, jarray y);           // l_cuhre.cpp
extern char* vadapt;                                 // l_quadvav.cpp
extern char* quadvav;                                // l_quadvav.cpp
jarray memo(monad um, dyad ud, jarray x, jarray y);  // l_cache.cpp
jarray memo(monad um, dyad ud, jarray y);            // l_cache.cpp
jarray gnuplot(jarray x, jarray y);                  // l_gnuplot.cpp
jarray gnuplot(jarray y);                            // l_gnuplot.cpp
jarray rasterplot(jarray x, jarray y);               // l_pnmplot.cpp
jarray pnmplot(jarray x, jarray y);                  // l_pnmplot.cpp
jarray pnmplot(jarray y);                            // l_pnmplot.cpp
jarray loadtrj(jarray x, jarray y);                  // l_loadtrj.cpp
jarray loadtrj(jarray y);                            // l_loadtrj.cpp

/** initializes yacts library functions and makes them avaliable to J */
void yacts::libInit() {
  defVerb("fft", fft, fft);
  defVerb("fftr2c", fftr2c, NULL);
  defVerb("fftc2r", NULL, fftc2r);
  defAdverb("cuba", cuba, cuba);
  defScript("vadapt",2,vadapt);
  defScript("quadvav",1,quadvav);
  string v_int =
    "1e_10 1e_5 u int y" "\n"
    ":" "\n"
    "if. 1 = # $ y do." "\n" // for 1-D integrals we use quadvav
    "{. x u quadvav y" "\n"  // we also strip the error estimate
    "else."                  // for n-D integrals we use cuhre
    "x u cuba y" "\n"
    "end." "\n";
  defScript("int",1,v_int);
  defAdverb("memo", memo, memo);
  defVerb("gnuplot", gnuplot, gnuplot);
  defVerb("rasterplot", NULL, rasterplot);
  defVerb("pnmplot", pnmplot, pnmplot);
  defVerb("loadtrj", loadtrj, loadtrj);
};

// readline inclusion for REPL
#include "config.h"
#include "cstdlib"
#ifdef HAVE_WORDEXP_H
#  include <wordexp.h>
#endif
#ifdef HAVE_LIBREADLINE
#  if defined(HAVE_READLINE_READLINE_H)
#    include <readline/readline.h>
#  elif defined(HAVE_READLINE_H)
#    include <readline.h>
#  else /* !defined(HAVE_READLINE_H) */
extern char *readline ();
#  endif /* !defined(HAVE_READLINE_H) */
char *cmdline = NULL;
#else /* !defined(HAVE_READLINE_READLINE_H) */
/* no readline */
#include <cstdio>
#define rl_abort
#define rl_bind_key(x,y)
#define add_history(x)
#define write_history(x)
#define read_history(x)

/** A simple emulation of GNU readline in case it is absent. 
    But to have real readline (especially with history) is so much better ! */
char* readline(string prompt) {
#define BUFSIZE 300
  char *buf=(char*) malloc(BUFSIZE);
  cout << prompt;
  if (cin.getline(buf, BUFSIZE, '\n'))
    return buf;
  else {
    free(buf);
    return NULL;
  };
};

#endif /* HAVE_LIBREADLINE */

#ifdef HAVE_READLINE_HISTORY
#  if defined(HAVE_READLINE_HISTORY_H)
#    include <readline/history.h>
#  elif defined(HAVE_HISTORY_H)
#    include <history.h>
#  else /* !defined(HAVE_HISTORY_H) */
#define add_history(x)
#define write_history(x)
#define read_history(x)
#  endif /* defined(HAVE_READLINE_HISTORY_H) */
         /* no history */
#endif /* HAVE_READLINE_HISTORY */

/** enters read(stdin)-eval-print(stdout) loop until the end of stdin */
int yacts::REPL() {
  // Before entering the loop we execute all meaningful sentences of the script
  vector<string> sentences=getProgram();
  for(int i=0;i<sentences.size();i++) {
    cout << "   " << sentences[i] << endl;
    doJ(sentences[i]);
  };

  // This is the REPL itself
  char *buf;
#ifdef HAVE_WORDEXP_H
  wordexp_t exp_history_name;
  wordexp("~/.yacts_history", &exp_history_name, 0);
  assert(exp_history_name.we_wordc==1);
  read_history(exp_history_name.we_wordv[0]);
#else
  read_history(".yacts_history");
#endif
  rl_bind_key('\t',rl_abort);
  while((buf = readline("   "))!=NULL) {
    doJ(buf);
    if (buf[0]!=0)
      add_history(buf);
  }
  printf("\b\b\b");
  free(buf);
#ifdef HAVE_WORDEXP_H
  write_history(exp_history_name.we_wordv[0]);
  wordfree(&exp_history_name);
#else
  write_history(".yacts_history");
#endif
  
  return 0;
};
